/* Copyright (c) 2010 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/* Publishing API.
 *
 * Provides useful methods and functions relating to publishing.
 */

ko.publishing = {};

(function() {

    XPCOMUtils.defineLazyGetter(Services, "koPublishingSvc", () =>
        Cc["@activestate.com/koPublishingService;1"].
        getService(Ci.koIPublishingService));

    var _publishing_configs = null;
    
    this.getConfigurations = function ko_pub_getConfigurations()
    {
        if (!_publishing_configs) {
            var settingsObj = {};
            Services.koPublishingSvc.getPublishingSettings({}, settingsObj);
            _publishing_configs = settingsObj.value;
        }
        return _publishing_configs;
    };

    this.saveConfigurations = function ko_pub_saveConfigurations(configs)
    {
        Services.koPublishingSvc.savePublishingSettings(configs.length, configs);
        // Drop the local publishing configs, as we need to re-fetch from
        // publishing service in order to have them properly sorted.
        _publishing_configs = null;
    };

    this.newConfiguration = function ko_pub_newConfiguration()
    {
        var args = {
            "type": "new"
        };
        ko.windowManager.openDialog('chrome://publishing/content/publish-dialog.xul',
                                    '_blank',
                                    'chrome,titlebar,resizable=yes',
                                    args);
    };

    this.synchronize = function ko_pub_synchronize(id)
    {
        var settings = null;
        var configs = ko.publishing.getConfigurations();
        for (var i=0; i < configs.length; i++) {
            if (configs[i].id == id) {
                settings = configs[i];
                break;
            }
        }
        var args = {
            "type": "synchronize",
            "settings": settings
        };
        ko.windowManager.openDialog('chrome://publishing/content/publish-dialog.xul',
                                    '_blank',
                                    'chrome,titlebar,resizable=yes',
                                    args);
    };

    this.onMenuPopupShowing = function ko_pub_onMenuPopupShowing(e)
    {
        var menupopup = e.originalTarget;
        if ( ! menupopup || menupopup.getAttribute("id") != "tools_publishing_menupopup")
            menupopup = document.getElementById("tools_publishing_menupopup");
            
        var separator = menupopup.getElementsByAttribute("id", "tools_publishing_newconfig_separator")[0];
        var prevSibling;
        var menuitem = separator.previousSibling;
        // Remove any existing menu items added previously.
        while (menuitem) {
            prevSibling = menuitem.previousSibling;
            if (menuitem.hasAttribute("dynamically_added_menuitem")) {
                menupopup.removeChild(menuitem);
            }
            menuitem = prevSibling;
        }
        var attributes;
        var configs = this.getConfigurations();
        for (var i=0; i < configs.length; i++) {
            attributes = {
                "label": configs[i].name,
                "oncommand": "ko.publishing.synchronize('" + configs[i].id + "')",
                "dynamically_added_menuitem": "true"
            }
            menuitem = xtk.domutils.newElement("menuitem", attributes);
            menupopup.insertBefore(menuitem, separator);
        }
    };

    this.onPlacesPopupShowing = function ko_pub_onPlacesPopupShowing(menupopup)
    {
        var enableMenus = false;
        var koFileExes = ko.places.manager.getSelectedFiles();
        for (var i=0; i < koFileExes.length; i++) {
            if (koFileExes[i].publishingStatus) {
                enableMenus = true;
                break;
            }
        }
        for (var i=0; i < menupopup.childNodes.length; i++) {
            if (enableMenus) {
                menupopup.childNodes[i].removeAttribute("disabled");
            } else {
                menupopup.childNodes[i].setAttribute("disabled", "true");
            }
        }
    };

        /**
         * Return the first publishing settings that match the given URI.
         * 
         * @param {string} uri - The file uri to match.
         * @returns {Components.interfaces.koIPublishingSettings}
         */
    this.getSettingsForUri = function ko_pub_getSettingsForUri(uri) {
        var configs = ko.publishing.getConfigurations();
        for (var i=0; i < configs.length; i++) {
            if (configs[i].matchesUri(uri)) {
                return configs[i];
            }
        }
        return null;
    };

        /**
         * Launches the diff dialog showing the file changes between the uris
         * and their corresponding local/remote publishing uris.
         *
         * @param {array} uris - The list of local or remote publishing uris.
         */
    this.diffURIs = function ko_pub_diffCurrentFile(uris) {
        var diffSvc = Components.classes["@activestate.com/koDiffService;1"].
                        getService(Components.interfaces.koIDiffService);
        var uri;
        var left_uris = [];
        var right_uris = [];

        for (var i=0; i < uris.length; i++) {
            uri = uris[i];
            var config = ko.publishing.getSettingsForUri(uri);
            if (!config) {
                continue;
            }
    
            var leftURI;
            var rightURI = uri;
    
            if (rightURI.substr(0, 5) == "file:") {
                // It's the local URI, work out the remote uri.
                leftURI = config.remote_uri + rightURI.substr(config.local_uri.length);
            } else {
                leftURI = config.local_uri + rightURI.substr(config.remote_uri.length);
            }
            left_uris.push(leftURI);
            right_uris.push(rightURI);
        }
        if (!left_uris) {
            var message = "No publishing configs found for any of: " + uris;
            require("notify/notify").send(msg, "publishing");
            return;
        }

        var diffContent = diffSvc.diffMultipleURIs(left_uris, left_uris.length,
                                                   right_uris, right_uris.length);
        var title = "Publishing Diff";
        ko.launch.diff(diffContent, title);
    }

    /**
     * Push the given uri.
     *
     * @param {String} uri
     */
    this.push = function publishing_push(uri) {
        // Implement the koIRemoteTransferCallback interface.
        var asyncHandler = {
            "callback": function(result, message) {
                if (result != Components.interfaces.koIAsyncCallback.RESULT_SUCCESSFUL) {
                    if (message.substr(0, 9) == "Conflict:") {
                        var fileSvc = Components.classes["@activestate.com/koFileService;1"].
                                        getService(Components.interfaces.koIFileService);
                        var koFileEx = fileSvc.getFileFromURI(uri);
                        koFileEx.publishingStatus = Components.interfaces.koISynchronizationCallback.SYNC_CONFLICT_BOTH_MODIFIED;
                        Services.obs.notifyObservers(null, "file_status", koFileEx.URI);
                    }
                    require("notify/notify").send(message, "publishing");
                }
            },
            "onProgress": function(self, message, value) {},
            "notifyFileTransferStarting": function(self, local_uri, remote_uri) {},
            "notifyFileTransferCompleted": function(self, local_uri, remote_uri) {},
            "notifyFileTransferFailed": function(self, local_uri, remote_uri, message) {},
        }
        Services.koPublishingSvc.pushLocalUri(uri, asyncHandler);
    }

    /************************************************************
     *         Controller for the publishing commands           *
     ************************************************************/

    function PublishingController() {
        ko.main.addWillCloseHandler((function() {
            window.controllers.removeController(this);
        }).bind(this));
    }
    PublishingController.prototype = new xtk.Controller();
    PublishingController.prototype.constructor = PublishingController;

    // Diff the current file.
    PublishingController.prototype.is_cmd_publishingDiffCurrentFile_enabled = function () {
        var view = ko.views.manager.currentView;
        var file = (view && view.koDoc && view.koDoc.file) || null;
        return file && file.publishingStatus;
    };
    PublishingController.prototype.do_cmd_publishingDiffCurrentFile = function () {
        var view = ko.views.manager.currentView;
        var file = (view && view.koDoc && view.koDoc.file) || null;
        if (!file) {
            var msg = "The current view does not reference a file";
            require("notify/notify").send(msg, "publishing");
            return;
        }
        ko.publishing.diffURIs([file.URI]);
    };

    // Push the current file.
    PublishingController.prototype.is_cmd_publishingPushCurrentFile_enabled = function () {
        var view = ko.views.manager.currentView;
        var file = (view && view.koDoc && view.koDoc.file) || null;
        return file && file.publishingStatus;
    };
    PublishingController.prototype.do_cmd_publishingPushCurrentFile = function () {
        var view = ko.views.manager.currentView;
        /** @type {Components.interfaces.koIFileEx} */
        var file = (view && view.koDoc && view.koDoc.file) || null;
        ko.publishing.push(file.URI);
    };

    // Create new configuration.
    PublishingController.prototype.is_cmd_publishingNewConfiguration_enabled = function () {
        return true;
    };
    PublishingController.prototype.do_cmd_publishingNewConfiguration = function () {
        ko.publishing.newConfiguration();
    };

    // Diff the places file(s).
    PublishingController.prototype.is_cmd_publishingPlacesDiffFile_enabled = function () {
        return true;
    };
    PublishingController.prototype.do_cmd_publishingPlacesDiffFile = function () {
        var uris = ko.places.manager.getSelectedUris();
        if (!uris.length) {
            return;
        }
        ko.publishing.diffURIs(uris);
    };

    // Push the places file(s).
    PublishingController.prototype.is_cmd_publishingPlacesPushFile_enabled = function () {
        return true;
    };
    PublishingController.prototype.do_cmd_publishingPlacesPushFile = function () {
        var uris = ko.places.manager.getSelectedUris();
        if (!uris.length) {
            return;
        }
        for (var i=0; i < uris.length; i++) {
            ko.publishing.push(uris[i]);
        }
    };


    /************************************************************
     *         Observer for catching configuration changes      *
     ************************************************************/

    this.observe = function ko_pub_observer(subject, topic, data) {
        if (topic == "publishing_configurations_changed") {
            _publishing_configs = null;
        }
    };

    this.initialize = function ko_pub_initialize() {
        // Controller instance on the current window to handle profiler commands.
        var pubController = new PublishingController();
        window.controllers.appendController(pubController);
        // Listen for pref changes.
        Services.obs.addObserver(ko.publishing, "publishing_configurations_changed", false);
    }

}).apply(ko.publishing);

window.addEventListener("komodo-ui-started", ko.publishing.initialize);
